/**
  *  \file client/dialogs/multitransfer.cc
  *  \brief Multi-Unit Cargo Transfer
  *
  *  This is a comparatively straightforward port of PCC 1.x's
  *  transfer.pas::DistributeCargo.
  *
  *  Multi-unit cargo transfer can transfer one type of cargo between
  *  multiple units. To support that, there are the following
  *  specialties:
  *  - hold space. Users unload to hold space, load from hold space.
  *    In PCC 1.x, hold space ist just an integer; here, it is a
  *    GHoldSpace object that normally takes part in cargo transfer.
  *  - temporary space. Users can add up to five of these. They are,
  *    too, implemented as GHoldSpace objects.
  *  - tagged object (extension). Essentially, acts as an extension
  *    to hold space; stuff taken when hold is empty is taken from
  *    here.
  *
  *  We set up a GCargoTransfer object with all units. The first one
  *  (index 0) always is hold space, followed by the ships, the planet,
  *  and finally the temporary spaces, if present.
  *
  *  Transfer can be confirmed only when all GHoldSpace (i.e. hold and
  *  temporary) are empty; this is verified by GCargoTransfer. In
  *  addition, we allow confirmation if hold space is not empty, but
  *  can be unloaded to the tagged object.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/multitransfer.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "game/cargo.h"
#include "game/cargotr.h"
#include "game/cargotype.h"
#include "game/global.h"
#include "game/objlist.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/unicodechars.h"

namespace {
    /** Hold space. A simple cargo container which can hold anything
        temporarily, but not permanently. */
    class GHoldSpace : public GCargoContainer {
        string_t name;
     public:
        GHoldSpace(string_t name);
        ~GHoldSpace();
        int32_t getCargoOnObject(GCargoType type);
        void changeCargoOnObject(GCargoType type, int32 amount);
        void changeReservedCargo(GCargoType type, int32 amount);
        bool canHaveCargo(GCargoType cargo_id);
        int32_t getMaxCargo(GCargoType cargo_id);
        int32_t getDisplayOffset(GCargoType cargo_id);
        string_t getName() const;
    };

    enum {
        // Layout, in ems
        UnitWidth = 15,
        HaveWidth = 6,
        FreeWidth = 6,
        HaveX = UnitWidth + HaveWidth,      // right justified
        FreeX = HaveX + FreeWidth,          // right justified
        TotalWidth = FreeX,

        // Pixels
        RightMarginPixels = 5
    };

    enum {
        cm_DistributeMenu = 2000
    };

    /** Multi-Unit Cargo Transfer: Header. Static table header widget. */
    class WMultiTransferHeader : public UIWidget {
     public:
        WMultiTransferHeader();
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
    };

    /** Multi-Unit Cargo Transfer: Hold space info. Displays content
        of hold space, and contains the "U" and "#" buttons. */
    class WMultiTransferHoldInfo : public UIBaseComplexWidget {
        SignalConnection conn_change;
        GCargoTransfer& tx;
        GCargoType what;
        UIButton btn_u;
        UIButton btn_menu;
     public:
        WMultiTransferHoldInfo(GCargoTransfer& tx, GCargoType what);
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
        void onResize(GfxRect oldSize, bool byLayout);
    };

    /** Multi-Unit Cargo Transfer: Unit list. Displays all
        participating units and implements all controls. */
    class WMultiTransferList : public UIListbox {
        SignalConnection conn_change;      /**< Connection to tx.sig_changed. */
        GCargoTransfer& tx;                /**< The cargo transfer. */
        GCargoType what;                   /**< What we are transferring. */
        int extension;                     /**< Tagged object. 0=none, otherwise index into tx. */
        int num_temps;                     /**< Number of temporary spaces (not counting hold space). */
        enum DistMode {
            distEqually,
            distFree,
            distProportionally
        };

        void onMenu();
     public:
        WMultiTransferList(GCargoTransfer& tx, GCargoType what, int extension);
        ~WMultiTransferList();

        void drawPart(GfxCanvas& can, int from, int to);
        bool handleEvent(const UIEvent& event, bool second_pass);
        void unload();
        void move(int32_t n);
        void distribute(DistMode mode);
        void doDistributeMenu(UIBaseWidget& anchor, bool context);

        bool isValid();
        void finish();

        GCargoTransfer& getTransfer()
            { return tx; }
    };

    /** Multi-Unit Cargo Transfer: Unit info. Displays some
        information about the currently-selected unit. */
    class WMultiTransferUnitInfo : public UIWidget {
        SignalConnection conn_change;
        SignalConnection conn_scroll;
        WMultiTransferList& list;
     public:
        WMultiTransferUnitInfo(WMultiTransferList& list);
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
    };

    /** Multi-Unit Cargo Transfer: Window. Implements dialog set-up
        and confirmation. */
    class WMultiTransferWindow : public UIWindow {
        ObjectHolder h;
        WMultiTransferList list;
        UIWidget* btn_ok;

        void init(GCargoTransfer& tx, GCargoType what);
        void onUpdateStatus();
        void onOK();
     public:
        WMultiTransferWindow(GCargoTransfer& tx, GCargoType what, int extension);
        void scrollTo(int n)
            { list.scrollTo(n-1, 1); }
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
    };
}

/******************************* GHoldSpace ******************************/

/** Constructor.
    \param name User-visible name of this object */
GHoldSpace::GHoldSpace(string_t name)
    : name(name)
{
    setFlags(can_Temporary);
}

GHoldSpace::~GHoldSpace()
{ }

int32_t
GHoldSpace::getCargoOnObject(GCargoType /*type*/)
{
    // Hold space always is empty.
    return 0;
}

void
GHoldSpace::changeCargoOnObject(GCargoType /*type*/, int32 amount)
{
    // We cannot change the content of hold space
    ASSERT(amount == 0);
}

void
GHoldSpace::changeReservedCargo(GCargoType /*type*/, int32 amount)
{
    // Since hold space is always empty, we cannot reserve anythign
    ASSERT(amount == 0);
}

bool
GHoldSpace::canHaveCargo(GCargoType /*cargo_id*/)
{
    // Hold space can hold anything
    return true;
}

int32_t
GHoldSpace::getMaxCargo(GCargoType /*cargo_id*/)
{
    // Hold space can hold ANYTHING
    return 1000000000;
}

int32_t
GHoldSpace::getDisplayOffset(GCargoType /*cargo_id*/)
{
    // No adjustment
    return 0;
}

string_t
GHoldSpace::getName() const
{
    return name;
}

/************************** WMultiTransferHeader *************************/

WMultiTransferHeader::WMultiTransferHeader()
{
    setState(st_Disabled, true);
}

void
WMultiTransferHeader::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());

    const int x = getExtent().x;
    const int y = getExtent().y;
    const int em = font_ems[FONT_NORMAL];

    outText(ctx, x, y, _("Unit"));
    ctx.setTextAlign(2, 0);
    outText(ctx, x + HaveX*em, y, _("Have"));
    outText(ctx, x + FreeX*em, y, _("Free"));

    can.update(getExtent());
}

void
WMultiTransferHeader::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(TotalWidth, 1).movedBy(RightMarginPixels, 0);
    info.flags = 0;
}

/************************* WMultiTransferHoldInfo ************************/

WMultiTransferHoldInfo::WMultiTransferHoldInfo(GCargoTransfer& tx, GCargoType what)
    : conn_change(observe(tx.sig_changed)),
      tx(tx),
      what(what),
      btn_u("U", 'u'),
      btn_menu(cm_DistributeMenu, "#", '#')
{
    addChildWidget(btn_u, 0);
    addChildWidget(btn_menu, 0);
    setState(st_Disabled, true);
    btn_menu.addCommand();
}

void
WMultiTransferHoldInfo::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(UIColor::tc_Static);
    ctx.setSolidBackground();
    outTextF(ctx, getExtent().x, getExtent().y,
             getExtent().w - 2*20 - 3,
             format(_("On hold: %d %s"),
                    tx.getContainer(0).getCargo(what),
                    getCargoUnit(what)));
    can.update(getExtent());
}

void
WMultiTransferHoldInfo::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_TITLE].scaledBy(15, 1);
    info.flags = li_FlexibleH;
}

void
WMultiTransferHoldInfo::onResize(GfxRect /*oldSize*/, bool /*byLayout*/)
{
    const int x = getExtent().getRightX();
    const int y = getExtent().y;
    const int w = 20;

    btn_menu.setExtent(GfxRect(x-w, y, w, w), true);
    btn_u.setExtent(GfxRect(x-2*w-3, y, w, w), true);
}

/*************************** WMultiTransferList **************************/

/** Constructor.
    \param tx Cargo transfer object to control
    \param what Cargo type to transfer
    \param extension Initial extension (tag) selection, index into tx, 0 means none
                     (0th object is hold space and cannot be extension) */
WMultiTransferList::WMultiTransferList(GCargoTransfer& tx, GCargoType what, int extension)
    : UIListbox(0,
                tx.getNumContainers() < 6 ? 5 :
                tx.getNumContainers() > 20 ? 20 : tx.getNumContainers()-1,
                TotalWidth*font_ems[FONT_NORMAL] + RightMarginPixels,
                tx.getNumContainers()-1),
      conn_change(observe(tx.sig_changed)),
      tx(tx),
      what(what),
      extension(extension),
      num_temps(0)
{
    sig_menu.add(this, &WMultiTransferList::onMenu);
}

WMultiTransferList::~WMultiTransferList()
{ }

/** Callback: context menu. */
void
WMultiTransferList::onMenu()
{
    doDistributeMenu(*this, true);
}

void
WMultiTransferList::drawPart(GfxCanvas& can, int from, int to)
{
    const int em = font_ems[FONT_NORMAL];

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    for (int i = from; i < to; ++i) {
        // Prepare item
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        // Draw item
        int line = getTop() + i + 1;    /* 1-based display, 0 is hold space! */
        if (line < tx.getNumContainers()) {
            // Container
            GCargoContainer& cnt = tx.getContainer(line);

            // Tag mark
            string_t text;
            if (line == extension) {
                text += UTF_BULLET;
            } else {
                text += " ";
            }

            // Name
            text += cnt.getName();
            ctx.setTextAlign(0, 0);
            outTextF(ctx, area.x, area.y, UnitWidth*em, text);

            // Have/Free
            int32_t now = cnt.getCargo(what);
            int32_t max = cnt.getMaxCargo(what);
            ctx.setTextAlign(2, 0);
            outTextF(ctx, area.x + HaveX*em, area.y, HaveWidth*em, numToString(now));

            // Free
            ctx.setColor(UIColor::tc_Faded);
            outTextF(ctx, area.x + FreeX*em, area.y, FreeWidth*em, max >= 20000 ? string_t(_("(unl)")) : numToString(max - now));
        }
    }
}

bool
WMultiTransferList::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case ' ':
         case '.':
            /* Move tag mark here, or untag. */
            {
                // Note 1-based indexing!
                int nextn = getCurrent()+1;
                if (nextn == extension) {
                    extension = 0;
                } else {
                    extension = nextn;
                }
                // Notify all listeners
                tx.sig_changed.raise();
            }
            return true;

         case SDLK_INSERT:
         case 'a':
            /* Create temporary */
            if (num_temps < 5) {
                if (num_temps != 0
                    || UIMessageBox(_("PCC2 will now create a temporary cargo storage which you can use "
                                      "to shuffle cargo around. It must be empty before you can finish "
                                      "this cargo transfer.\n"
                                      "Proceed?"),
                                    _("Cargo Transfer")).doYesNoDialog())
                {
                    ++num_temps;
                    tx.addNewContainer(new GHoldSpace(format(_("Temporary Storage #%d"), num_temps)));
                    setSize(tx.getNumContainers()-1);
                    scrollTo(tx.getNumContainers()-1, 1);
                }
            }
            return true;

         case 'u':
            /* Unload everything to hold */
            unload();
            return true;

         case 'e':
            /* Distribute equally */
            distribute(distEqually);
            return true;

         case 'f':
            /* Distribute equal free space */
            distribute(distFree);
            return true;

         case 'p':
            /* Distribute proportionally */
            distribute(distProportionally);
            return true;

         case SDLK_LEFT:
         case '+':
            move(event.key.arg ? event.key.arg : 10);
            return true;

         case SDLK_LEFT + ss_Shift:
         case '+' + ss_Shift:
            move(1);
            return true;

         case SDLK_LEFT + ss_Ctrl:
         case '+' + ss_Ctrl:
            move(100);
            return true;

         case SDLK_LEFT + ss_Alt:
         case '+' + ss_Alt:
            move(0x7FFFFFFF);
            return true;

         case SDLK_RIGHT:
         case '-':
            move(event.key.arg ? -event.key.arg : -10);
            return true;

         case SDLK_RIGHT + ss_Shift:
         case '-' + ss_Shift:
            move(-1);
            return true;

         case SDLK_RIGHT + ss_Ctrl:
         case '-' + ss_Ctrl:
            move(-100);
            return true;

         case SDLK_RIGHT + ss_Alt:
         case '-' + ss_Alt:
            move(-0x7FFFFFFF);
            return true;
        }
    }
    return UIListbox::handleEvent(event, second_pass);
}

/** Unload everything to hold space. */
void
WMultiTransferList::unload()
{
    for (int i = 1, n = tx.getNumContainers(); i < n; ++i) {
        if (i != extension) {
            tx.moveFT(what, tx.getContainer(i).getCargo(what), i, 0, true, false);
        }
    }
}

/** Move cargo.
    \param n amount to add to current unit. Negative to remove. */
void
WMultiTransferList::move(int32_t n)
{
    int me = getCurrent()+1;
    if (n > 0) {
        // Add to this unit, and take from hold space.
        if (tx.moveFT(what, n, 0, me, true, false) == 0) {
            // Hold space was empty. Anything in extension?
            if (extension != 0 && extension != me) {
                tx.moveFT(what, n, extension, me, true, false);
            }
        }
    } else {
        // Remove from unit and add to hold space. This always succeeds.
        tx.moveFT(what, -n, me, 0, true, false);
    }
}

/** Distribute cargo. Implements the 'e', 'f', and 'p' cargo
    distribution modes.

    All modes are implemented using a "try until everything
    distributed" loop, to clean up rounding leftovers.

    \param mode Mode */
void
WMultiTransferList::distribute(DistMode mode)
{
    const int32_t MAX_UNIT_CARGO = 20000;
    int paranoiaCounter = 100;
    while (1) {
        // Check main parameters
        if (--paranoiaCounter < 0) {
            break;
        }

        int32_t holdAmount = tx.getContainer(0).getCargo(what);
        if (holdAmount == 0) {
            break;
        }

        // Count units
        int nUnitsWithRoom = 0;
        int32_t totalFree = 0;
        int32_t totalRoom = 0;
        for (int i = 1, n = tx.getNumContainers(); i < n; ++i) {
            GCargoContainer& cnt = tx.getContainer(i);
            if (i != extension && !cnt.hasFlag(cnt.can_Temporary)) {
                int32_t max = std::min(cnt.getMaxCargo(what), MAX_UNIT_CARGO);
                int32_t have = cnt.getCargo(what);
                if (have < max) {
                    ++nUnitsWithRoom;
                    totalFree += max - have;
                    totalRoom += max;
                }
            }
        }
        if (nUnitsWithRoom == 0) {
            break;
        }

        // Distribute
        int32_t each, total;
        switch (mode) {
         case distEqually:
            // Distribute equally. Try to load ceil(holdAmount/nUnitsWithRoom) to each.
            each = (holdAmount + nUnitsWithRoom-1) / nUnitsWithRoom;
            for (int i = 1, n = tx.getNumContainers(); i < n; ++i) {
                GCargoContainer& cnt = tx.getContainer(i);
                if (i != extension && !cnt.hasFlag(cnt.can_Temporary)) {
                    tx.moveFT(what, each, 0, i, true, false);
                }
            }
            break;

         case distFree:
            // Distribute free space. Try to make each unit have
            // (totalFree-holdAmount)/nUnitsWithRoom free space.
            if (holdAmount > totalFree) {
                each = 0;
            } else {
                each = (totalFree - holdAmount) / nUnitsWithRoom;
            }
            for (int i = 1, n = tx.getNumContainers(); i < n; ++i) {
                GCargoContainer& cnt = tx.getContainer(i);
                if (i != extension && !cnt.hasFlag(cnt.can_Temporary)) {
                    int32_t free = cnt.getMaxCargo(what) - cnt.getCargo(what);
                    if (free > each) {
                        tx.moveFT(what, free - each, 0, i, true, false);
                    }
                }
            }
            break;

         case distProportionally:
            // Distribute proportional. We want each ship to have
            // cargo proportional to its total capacity, i.e. it
            // should have capacity*totalCargo/totalRoom. Some units
            // max already have more than that, so they don't receive
            // anything here; we'll need multiple iterations to sort
            // this out.
            total = totalRoom - totalFree + holdAmount;
            for (int i = 1, n = tx.getNumContainers(); i < n; ++i) {
                GCargoContainer& cnt = tx.getContainer(i);
                if (i != extension && !cnt.hasFlag(cnt.can_Temporary)) {
                    int32_t max = std::min(cnt.getMaxCargo(what), MAX_UNIT_CARGO);
                    int32_t want = (total * max + (totalRoom-1)) / totalRoom;
                    int32_t have = cnt.getCargo(what);
                    if (want > have) {
                        tx.moveFT(what, want - have, 0, i, true, false);
                    }
                }
            }
            break;

         default:
            paranoiaCounter = 0;
            break;
        }
    }
}

/** Distribute (context) menu.
    \param anchor Anchor widget for menu popup
    \param context true iff this is the context menu, false if this is the button menu */
void
WMultiTransferList::doDistributeMenu(UIBaseWidget& anchor, bool context)
{
    StringList list;
    if (tx.getContainer(0).getCargo(what) != 0) {
        // Distributions only when hold is nonempty
        list.add(distEqually, _("E - Add Equal"));
        list.add(distFree, _("F - Equal Free Space"));
        list.add(distProportionally, _("P - Proportional"));
    }
    list.add(1000, _("U - Unload"));
    if (context) {
        // Context menu: offer tag/untag
        if (getCurrent()+1 == extension) {
            list.add(1001, _("Space - Un-tag"));
        } else {
            list.add(1002, _("Space - Tag"));
        }
    }

    UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    if (box.doStandardMenu(anchor)) {
        switch (box.getKey()) {
         case 1000:
            unload();
            break;
         case 1001:
            extension = 0;
            tx.sig_changed.raise();
            break;
         case 1002:
            extension = getCurrent()+1;
            tx.sig_changed.raise();
            break;
         default:
            distribute(DistMode(box.getKey()));
            break;
        }
    }
}

/** Check for validity. This returns true when a call to finish()
    would bring the GCargoTransfer into a state where it can be
    committed.

    This is a little diffent test than in GCargoTransfer, because it
    honors the implicit hold extension. */
bool
WMultiTransferList::isValid()
{
    // All objects must be valid
    for (int i = 1, n = tx.getNumContainers(); i < n; ++i) {
        GCargoContainer& cnt = tx.getContainer(i);
        if (cnt.hasFlag(GCargoContainer::can_Temporary) && cnt.isChanged()) {
            // Temporary container with nonzero content
            return false;
        }
    }

    // Validate hold
    GCargoContainer& hold = tx.getContainer(0);
    if (hold.isChanged()) {
        // Hold is not empty. There must be an extension, which is not temporary,
        // and can hold our content.
        if (extension == 0)
            return false;
        GCargoContainer& ext = tx.getContainer(extension);
        if (ext.hasFlag(GCargoContainer::can_Temporary))
            return false;
        if (ext.getCargo(what) + hold.getCargo(what) > ext.getMaxCargo(what))
            return false;
    }

    // All tests passed.
    return true;
}

/** Finish dialog. Moves hold into the extension, in order to make it empty. */
void
WMultiTransferList::finish()
{
    int32_t hold = tx.getContainer(0).getCargo(what);
    if (hold != 0 && extension != 0) {
        tx.moveFT(what, hold, 0, extension, true, false);
    }
}

/************************* WMultiTransferUnitInfo ************************/

WMultiTransferUnitInfo::WMultiTransferUnitInfo(WMultiTransferList& list)
    : conn_change(observe(list.getTransfer().sig_changed)),
      conn_scroll(observe(list.sig_moved)),
      list(list)
{ }

void
WMultiTransferUnitInfo::drawContent(GfxCanvas& can)
{
    // Base information
    GCargoContainer& cnt = list.getTransfer().getContainer(list.getCurrent()+1);
    string_t line1 = cnt.getName();
    string_t line2;

    // Additional information
    if (GShipTransfer* shipx = dynamic_cast<GShipTransfer*>(&cnt)) {
        GShip& ship = shipx->getShip();
        line1 += format(": %s", ship.getHull().getShortName());
        if (ship.getBeamCount() > 0)
            line1 += format(", %d" UTF_TIMES "%s", ship.getBeamCount(), getBeam(ship.getBeamType()).getShortName());
        if (ship.getTorpLauncherCount() > 0)
            line1 += format(", %d" UTF_TIMES "%s", ship.getTorpLauncherCount(), getTorp(ship.getTorpType()).getShortName());
        if (ship.getBayCount() > 0)
            line1 += format(_(", %d" "\xC3\x97" "Ftr"), ship.getBayCount());
        line2 = format(_("FCode: \"%s\", Damage: %d%%"), ship.getFCode(), ship.getDamage());
    }

    // Show it
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Faded);
    ctx.setSolidBackground();

    const int x = getExtent().x;
    const int y = getExtent().y;
    const int w = getExtent().w;

    outTextF(ctx, x, y, w, line1);
    outTextF(ctx, x, y+font_heights[FONT_NORMAL], w, line2);

    can.update(getExtent());
}

void
WMultiTransferUnitInfo::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(15, 2);
    info.flags = li_FlexibleH;
}

/************************** WMultiTransferWindow *************************/

WMultiTransferWindow::WMultiTransferWindow(GCargoTransfer& tx, GCargoType what, int extension)
    : UIWindow(format(_("Transfer %s"), getCargoName(what)), BLUE_WINDOW, COLOR_GRAY),
      list(tx, what, extension)
{
    init(tx, what);
}

void
WMultiTransferWindow::init(GCargoTransfer& tx, GCargoType what)
{
    // VBox
    //   VBox [0]
    //     WMultiTransferHeader
    //     UIFrameGroup (HBox)
    //       list
    //       UIScrollbar
    //   WMultiTransferUnitInfo
    //   WMultiTransferHoldInfo
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer
    //     "Help"

    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIVBoxLayout::instance0));
    UIFrameGroup& g11 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup& g4 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g11.add(list);
    g11.add(h.add(new UIScrollbar(list)));
    g1.add(h.add(new WMultiTransferHeader()));
    g1.add(g11);
    add(g1);

    add(h.add(new WMultiTransferUnitInfo(list)));
    add(h.add(new WMultiTransferHoldInfo(tx, what)));

    btn_ok = &h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &WMultiTransferWindow::onOK);
    g4.add(*btn_ok);
    g4.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g4.add(h.add(new UISpacer()));
    g4.add(h.add(new UIButton(_("Help"), 'h')));
    add(g4);
    add(h.add(new WHelpWidget("pcc2:multicargo")));

    pack();

    tx.sig_changed.add(this, &WMultiTransferWindow::onUpdateStatus);
    onUpdateStatus();
}

void
WMultiTransferWindow::onUpdateStatus()
{
    btn_ok->setState(st_Disabled, !list.isValid());
}

void
WMultiTransferWindow::onOK()
{
    if (list.isValid()) {
        list.finish();
        stop(1);
    }
}

bool
WMultiTransferWindow::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_DistributeMenu) {
        list.doDistributeMenu(sender, false);
        return true;
    } else {
        return UIWindow::handleCommand(cmd, sender);
    }
}

/***************************** Dialog Set Up *****************************/

/** Dialog set up: ask for cargo type.
    \param unit [in] Unit that's providing the cargo
    \param what [out] Chosen cargo type
    \return true on ok, false on cancel */
static bool
askCargoType(GCargoContainer& unit, GCargoType& what)
{
    StringList list;
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i) {
        if (unit.canHaveCargo(i)) {
            list.add(i, getCargoName(i));
        }
    }
    UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list), list);
    if (box.doStandardListbox(_("Cargo Transfer"))) {
        what = GCargoType(box.getKey());
        return true;
    } else {
        return false;
    }
}

/** Dialog set up: add newly-constructed cargo container to transfer.
    \param tx [in/out] Cargo transfer
    \param what [in] Cargo type
    \param cnt [in] Newly-created cargo container
    \retval true Container was added to the transfer
    \retval false Container cannot contain desired cargo type and was deleted */
static bool
tryAdd(GCargoTransfer& tx, GCargoType what, GCargoContainer* cnt)
{
    if (cnt->canHaveCargo(what)) {
        tx.addNewContainer(cnt);
        return true;
    } else {
        delete cnt;
        return false;
    }
}

/** Build cargo transfer list.
    \param tx [out] List will be added to this object
    \param what [in] Cargo type to transfer
    \param ship [in] Invoking ship
    \param fleet [in] true to limit to fleet members
    \param thisShipSlot [out] Index into tx of container for this ship (=initial selection)
    \param initialExtension [out] Index into tx of container for planet (=initial extension) */
static void
buildTransferList(GCargoTransfer& tx, const GCargoType what, const GShip& ship, bool fleet, int& thisShipSlot, int& initialExtension)
{
    // Initialize
    thisShipSlot = 0;
    initialExtension = 0;

    // First object always is hold space
    tx.addNewContainer(new GHoldSpace(_("Hold space")));

    // Ships in order of 'l'
    GObjectList list;
    list.addObjectsAt(getDisplayedTurn().getCurrentUniverse(), ship.getPos(), 0, 0);
    list.postprocessUser(false);
    for (int i = 1, n = list.getNumObjects(); i <= n; ++i) {
        // 1-based access!
        GShip* s2 = dynamic_cast<GShip*>(&list.getObject(i));
        if (s2 != 0
            && s2->getOwner() == ship.getOwner()
            && (!fleet || s2->getFleetNumber() == ship.getFleetNumber()))
        {
            // Let's use this ship
            if (s2->getId() == ship.getId()) {
                thisShipSlot = tx.getNumContainers();
            }
            tryAdd(tx, what, new GShipTransfer(*s2));
        }
    }

    // Is there a planet?
    if (int pid = ship.getOrbit()) {
        GPlanet& p = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
        if (p.isPlayed() && p.getOwner() == ship.getOwner()) {
            // We play this planet, so use it.
            if (tryAdd(tx, what, new GPlanetTransfer(p))) {
                initialExtension = tx.getNumContainers()-1;
            }
        }
    }
}

/** Multi-Unit Cargo Transfer.
    \param sid Invoking ship
    \param fleet true to restrict to fleet */
void
doMultiTransfer(int sid, bool fleet)
{
    // Validate ship
    GShip& ship = getDisplayedTurn().getCurrentUniverse().getShip(sid);
    if (!ship.isPlayed()) {
        return;
    }

    // Find out what can be transferred. For simplicity, use a temporary
    // transaction to enumerate the types (PCC 1.x attempts to re-use
    // the object).
    GCargoType what;
    {
        GShipTransfer stx(ship);
        if (!askCargoType(stx, what)) {
            return;
        }
    }

    // Enumerate
    GCargoTransfer tx;
    int thisShipSlot;
    int initialExtension;
    buildTransferList(tx, what, ship, fleet, thisShipSlot, initialExtension);

    // Check availability of cargo
    int32_t totalCargo = 0;
    for (int i = 0, n = tx.getNumContainers(); i < n; ++i) {
        totalCargo += tx.getContainer(i).getCargo(what);
    }
    if (totalCargo == 0) {
        messageBox(format(_("Nobody has %s at this place."), getCargoName(what)),
                   _("Cargo Transfer"));
        return;
    }

    // Need at least 3 items: hold space and two units
    if (tx.getNumContainers() < 3) {
        if (fleet) {
            messageBox(format(_("There is no other fleet member at this place which could carry %s."),
                              getCargoName(what)),
                       _("Cargo Transfer"));
        } else {
            messageBox(format(_("There is no other unit at this place which could carry %s."),
                              getCargoName(what)),
                       _("Cargo Transfer"));
        }
        return;
    }

    // Build window
    WMultiTransferWindow window(tx, what, initialExtension);
    window.scrollTo(thisShipSlot);
    ui_root->centerWidget(window);
    if (ui_root->runModal(window)) {
        tx.doCommit();
    } else {
        tx.doCancel();
    }
}
