/**
  *  \file client/tasksupport.cc
  *  \brief Auto Task Support Functions
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tasksupport.h"
#include "cpluslib/format.h"
#include "game/obj.h"
#include "game/spec.h"
#include "game/struct.h"
#include "int/autotask.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/process.h"
#include "ui/commands.h"
#include "ui/keystring.h"
#include "ui/msgbox.h"
#include "util/nls.h"

/** Get auto task for an object.
    \param obj The object
    \param kind Kind of auto task (pkShipTask, pkPlanetTask, pkBaseTask)
    \param create true to create the task if it does not exist */
IntExecutionContext*
getAutoTaskForObject(GObject& obj, IntExecutionContext::ProcessKind kind, bool create)
{
    if (IntExecutionContext* ctx = getProcessByObject(&obj, kind)) {
        return ctx;
    } else if (create) {
        // Create process
        const char* fmt = (kind == IntExecutionContext::pkShipTask
                           ? _("Auto Task Ship %d")
                           : kind == IntExecutionContext::pkPlanetTask
                           ? _("Auto Task Planet %d")
                           : _("Auto Task Starbase %d"));
        IntExecutionContext& ctx = createProcess(format(fmt, obj.getId()));

        // Place it in planet context
        if (kind == IntExecutionContext::pkShipTask) {
            ctx.pushNewContext(new IntShipContext(obj.getId()));
        } else {
            ctx.pushNewContext(new IntPlanetContext(obj.getId()));
        }
        ctx.markContextTOS();

        // Mark as auto task
        ctx.setProcessKind(kind);
        ctx.setState(IntExecutionContext::psSuspended);
        return &ctx;
    } else {
        return 0;
    }
}

/** Add command to an auto task. It will ask the user where to place
    the command.
    \param task Task to add to
    \param cmd Command to add */
bool
addToAutoTask(IntExecutionContext* task, string_t cmd)
{
    // Can we modify this task?
    if (!task || task->getState() != IntExecutionContext::psSuspended) {
        messageBox(_("Unable to modify Auto Task at this point. "
                     "It might be in use by another part of the program."),
                   _("Auto Task"));
        return false;
    }

    // Edit the task. We ignore errors from loadFromProcess, like the
    // actual auto task screen currently does.
    IntAutoTaskEditor editor;
    editor.loadFromProcess(*task);
    task->setState(IntExecutionContext::psFrozen);

    try {
        enum { Before = 1, End = 2, Cancel = 3 };
        int action = End;
        if (editor.getPC() < editor.getNumInstructions()) {
            UIKeyString b(_("Before")), e(_("End"));
            UIMessageBox msg((format(_("This unit is already executing the command \"%s\". "
                                       "Do you want to execute the new order before that, "
                                       "or do you want it at the end of its current task?"),
                                     editor[editor.getPC()])),
                             _("Auto Task"));
            msg.addButton(Before, b.getString(), b.getKey());
            msg.addButton(End,    e.getString(), e.getKey());
            msg.addButton(Cancel, _("Cancel"), SDLK_ESCAPE);
            action = msg.run();
            if (action == cm_Enter)
                action = Before;
            if (action == cm_Escape)
                action = Cancel;
        }

        if (action == Before) {
            // Insert before PC
            uint32_t pos = editor.getPC();
            editor.replace(pos, 0, cmd);
            editor.setPC(pos);
        }
        if (action == End) {
            // Insert at end. Insert before any Restart command. Skip blanks, too.
            uint32_t pos = editor.getNumInstructions();
            while (pos > 0
                   && (editor.isRestartCommand(editor[pos-1])
                       || editor.isBlankCommand(editor[pos-1])))
            {
                --pos;
            }
            editor.replace(pos, 0, cmd);
        }

        if (action != Cancel) {
            // Task was modified
            editor.saveToProcess(*task);
            task->setState(IntExecutionContext::psRunnable);
            task->run();
            return true;
        } else {
            // Cancel, so don't modify the task
            task->setState(IntExecutionContext::psSuspended);
            return false;
        }
    }
    catch (...) {
        // If anything goes wrong, at least release the task
        task->setState(IntExecutionContext::psSuspended);
        throw;
    }
}

/** Create a "BuildShip"/"EnqueueShip" command. This is needed at multiple
    places, therefore it's here. */
string_t
makeBuildOrderCommand(string_t verb, const TBuildOrder& order)
{
    if (order.hull == 0) {
        return verb + " 0";
    } else {
        return format("%s %d, %d, %d, %d, %d, %d   %% %s")
            << verb
            << order.hull
            << order.engine
            << order.beam_type << order.beam_count
            << order.torp_type << order.torp_count
            << getHull(order.hull).getShortName();
    }
}
