/**
  *  \file int/if/ufoif.cc
  *  \brief Interpreter Game Interface: Ufo Properties
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/ufoif.h"
#include "cpluslib/format.h"
#include "game/tables.h"
#include "game/ufo.h"
#include "int/interror.h"
#include "int/value.h"
#include "int/if/ifutil.h"
#include "int/data.h"
#include "game/global.h"
#include "cpluslib/misc.h"
#include "int/compound.h"
#include "int/args.h"
#include "int/if/objif.h"
#include "int/propenum.h"

/** Implementation of "Ufo(nn)".
    \param args function parameters
    \return value representing "Ufo(nn)" */
IntValue*
IFUfoGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1, 1);
    if (!checkIntArg(id, args.getNext(), 0, 32767))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    GUfoContainer& con = getDisplayedTurn().getCurrentUniverse().getUfos();
    GUfoContainer::iterator it = con.getFirstAt(id);
    if (it != con.end() && it->getId() == id) {
        return new IntUfoContext(con, it);
    } else {
        return 0;
    }
}

/** Implementation of "ForEach Ufo".
    \return Context representing first Ufo */
IntContext*
IFUfoMake()
{
    if (!haveDisplayedTurn())
        return 0;

    GUfoContainer& con = getDisplayedTurn().getCurrentUniverse().getUfos();
    GUfoContainer::iterator it = con.begin();
    if (it != con.end()) {
        return new IntUfoContext(con, it);
    } else {
        return 0;
    }
}

/***************************** IntUfoContext *****************************/

namespace { enum { UfoProperty, UfoMethod }; }

static const TIntMapping ufo_mapping[] = {
    { "COLOR",          iupColorPCC,      UfoProperty, thInt },
    { "COLOR.EGA",      iupColorEGA,      UfoProperty, thInt },
    { "HEADING",        iupHeadingName,   UfoProperty, thString },
    { "HEADING$",       iupHeadingInt,    UfoProperty, thInt },
    { "ID",             iupId,            UfoProperty, thInt },
    { "ID2",            iupId2,           UfoProperty, thInt },
    { "INFO1",          iupInfo1,         UfoProperty, thString },
    { "INFO2",          iupInfo2,         UfoProperty, thString },
    { "KEEP",           iupKeepFlag,      UfoProperty, thBool },
    { "LASTSCAN",       iupLastScan,      UfoProperty, thInt },
    { "LOC.X",          iupLocX,          UfoProperty, thInt },
    { "LOC.Y",          iupLocY,          UfoProperty, thInt },
    { "MARK",           0,                UfoMethod,   thProcedure },
    { "MARKED",         iupMarked,        UfoProperty, thBool },
    { "MOVE.DX",        iupMoveDX,        UfoProperty, thInt },
    { "MOVE.DY",        iupMoveDY,        UfoProperty, thInt },
    { "NAME",           iupName,          UfoProperty, thString },
    { "RADIUS",         iupRadius,        UfoProperty, thInt },
    { "SPEED",          iupSpeedName,     UfoProperty, thString },
    { "SPEED$",         iupSpeedInt,      UfoProperty, thInt },
    { "TYPE",           iupType,          UfoProperty, thInt },
    { "UNMARK",         1,                UfoMethod,   thProcedure },
    { "VISIBLE.PLANET", iupVisiblePlanet, UfoProperty, thInt },
    { "VISIBLE.SHIP",   iupVisibleShip,   UfoProperty, thInt },
};

static const IntObjectProcedureValue::call_t ufo_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
};



IntUfoContext::IntUfoContext(GUfoContainer& con, GUfoContainer::iterator it)
    : con(con), it(it)
{ }

IntUfoContext::~IntUfoContext()
{ }

int32_t
IntUfoContext::lookup(const string_t& name)
{
    return lookupName(name, ufo_mapping, countof(ufo_mapping));
}

void
IntUfoContext::set(int32_t index, IntValue* value)
{
    switch (ufo_mapping[index].domain) {
     case UfoProperty:
        setUfoProperty(*it, IntUfoProperty(ufo_mapping[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntUfoContext::get(int32_t index)
{
    switch (ufo_mapping[index].domain) {
     case UfoProperty:
        return getUfoProperty(*it, IntUfoProperty(ufo_mapping[index].index));
     case UfoMethod:
        return new IntObjectProcedureValue(*it, ufo_methods[ufo_mapping[index].index]);
     default:
        return 0;
    }
}

bool
IntUfoContext::next()
{
    GUfoContainer::iterator it2 = it;
    ++it2;
    if (it2 != con.end()) {
        it = it2;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntUfoContext::clone()
{
    return new IntUfoContext(con, it);
}

GObject*
IntUfoContext::getObject()
{
    return &*it;
}

string_t
IntUfoContext::toString(bool /*readable*/)
{
    // FIXME?
    return "#<ufo>";
}

void
IntUfoContext::store(IntTagNode& sv, Stream& aux)
{
    // FIXME?
    throw IntError::notSerializable();
}

void
IntUfoContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(ufo_mapping, countof(ufo_mapping));
}

/** Get Ufo property.
    \param ufo Ufo to query
    \param iup Property to query */
IntValue*
getUfoProperty(const GUfo& ufo, IntUfoProperty iup)
{
    switch (iup) {
     case iupColorEGA:
        return makeIntValue(ufo.getColorCode());
     case iupColorPCC:
        return makeIntValue(ufo.getColor());
     case iupHeadingInt:
        if (ufo.getHeading() >= 0)
            return makeIntValue(ufo.getHeading());
        else
            return 0;
     case iupHeadingName:
        // FIXME: NLS
        if (ufo.getHeading() >= 0)
            return makeStringValue(getHeadingName(ufo.getHeading()));
        else
            return 0;
     case iupId:
        return makeIntValue(ufo.getId());
     case iupId2:
        return makeIntValue(ufo.getRealId());
     case iupInfo1:
        return makeStringValue(ufo.getInfo1());
     case iupInfo2:
        return makeStringValue(ufo.getInfo2());
     case iupKeepFlag:
        return makeBoolValue(ufo.isStoredInHistory());
     case iupLastScan:
        return makeIntValue(ufo.getLastTurn());
     case iupLocX:
        return makeIntValue(ufo.getPos().x);
     case iupLocY:
        return makeIntValue(ufo.getPos().y);
     case iupMarked:
        return makeBoolValue(ufo.isMarked());
     case iupMoveDX:
        return makeIntValue(ufo.getMovementVector().x);
     case iupMoveDY:
        return makeIntValue(ufo.getMovementVector().y);
     case iupName:
        return makeStringValue(ufo.getName());
     case iupRadius:
        return makeIntValue(ufo.getRadius());
     case iupSpeedInt:
        if (ufo.getSpeed() >= 0)
            return makeIntValue(ufo.getSpeed());
        else
            return 0;
     case iupSpeedName:
        if (ufo.getSpeed() >= 0)
            return makeStringValue(format("Warp %d", ufo.getSpeed()));
        else
            return 0;
     case iupType:
        return makeIntValue(ufo.getTypeCode());
     case iupVisiblePlanet:
        return makeIntValue(ufo.getPlanetRange());
     case iupVisibleShip:
        return makeIntValue(ufo.getShipRange());
    }
    return 0;
}

/** Set Ufo property.
    \param ufo Ufo to modify
    \param iup Property to modify
    \param value Value to set */
void
setUfoProperty(GUfo& ufo, IntUfoProperty iup, IntValue* value)
{
    (void) ufo;
    (void) iup;
    (void) value;

    throw IntError::notAssignable();
}
