/**
  *  \file int/mutex.h
  *  \brief Interpreter: Mutexes (Locks)
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_MUTEX_H
#define PCC_V2_INT_MUTEX_H

#include <vector>
#include "int/context.h"

class IntExecutionContext;

/** Mutex data. This holds the data associated with a mutex. These objects
    are reference-counted and can only be created using the create() function.
    All objects are stored in the global mutex_objects registry. */
class IntMutex {
    friend class IntMutexContext;

    /** Reference counter. When this goes to 0, the mutex is freed. */
    uint32_t reference_counter;
    /** Slot. Slot in mutex_objects registry. Corresponds to the name. */
    const uint32_t slot;
    /** Note. User's not associated with this mutex. */
    const string_t note;
    /** Owner. If non-null, the process owning this mutex. Otherwise, the
        mutex is orphaned. */
    IntExecutionContext* owner;

    IntMutex(uint32_t slot, const string_t& note, IntExecutionContext* owner);
    static IntMutex* createMutex(uint32_t slot, const string_t& note, IntExecutionContext* owner);

 public:
    ~IntMutex();

    IntMutex* addReference();
    void removeReference();

    const string_t& getNote() const;
    const string_t& getName() const;
    IntExecutionContext* getOwner() const;
    void setOwner(IntExecutionContext* newOwner);

    static IntMutex* create(const string_t& name, const string_t& note, IntExecutionContext* owner);
    static IntMutex* query(const string_t& name);
    static void disownLocksByProcess(IntExecutionContext* owner);

    static void enumMutexes(std::vector<IntMutex*>& data, IntExecutionContext* owner);
};

/** Mutex context. This is the main primitive exposed to the script interface.
    Users will do "With Lock(...)", causing an IntMutexContext be created.
    As long as this context lives, the mutex will be held. Since the interpreter
    may copy around the object, we must use reference-counting. */
class IntMutexContext : public IntContext {
    IntMutex* mtx;

 public:
    IntMutexContext(IntMutex* mtx);
    ~IntMutexContext();

    // IntContext:
    int32_t lookup(const string_t& name);
    void set(int32_t index, IntValue* value);
    IntValue* get(int32_t index);
    bool next();
    IntContext* clone();
    GObject* getObject();
    void enumProperties(IntPropertyAcceptor& acceptor);

    // IntValue:
    string_t toString(bool readable);
    void store(IntTagNode& sv, Stream& aux);

    static IntMutexContext* load(uint32_t flag, Stream& aux);
};

void initInterpreterMutexInterface();


/** Increase reference counter.
    \return this */
inline IntMutex*
IntMutex::addReference()
{
    ++reference_counter;
    return this;
}

/** Get user's note. */
inline const string_t&
IntMutex::getNote() const
{
    return note;
}

/** Get owning process.
    \return the process, or 0 if orphaned */
inline IntExecutionContext*
IntMutex::getOwner() const
{
    return owner;
}

/** Set owning process. Use with care. */
inline void
IntMutex::setOwner(IntExecutionContext*const newOwner)
{
    owner = newOwner;
}

#endif
